/*
 * base64.c
 *
 * This code was adapted from GNU Anubis, version 3.6.2
 * Copyright (C) 2001, 2002 The Anubis Team.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <limits.h>

#include "merror.h"
#include "xmalloc.h"


#ifndef USE_GSASL

/*
 * base64enc()
 *
 * Encode the data in 's' (length of the data is 'len') in BASE64. The returned
 * string should be freed when not used anymore.
 */

char *base64enc(char *s, size_t s_len)
{
    char b64chars[] = 
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
    char *t;
    char *p;
    int x, y;
    int len;
    
    if (s_len > (size_t)INT_MAX)
    {
	die("data too long in base64enc()");
    }
    len = (int)s_len;
    t = xmalloc((4 * ((len + 2) / 3) + 1) * sizeof(char));
    p = t;
    while (len-- > 0)
    {
	x = *s++;
	*p++ = b64chars[(x >> 2) & 63];
	if (len-- <= 0) 
	{
	    *p++ = b64chars[(x << 4) & 63];
    	    *p++ = '=';
	    *p++ = '=';
    	    break;
	}
	y = *s++;
	*p++ = b64chars[((x << 4) | ((y >> 4) & 15)) & 63];
	if (len-- <= 0) 
	{
	    *p++ = b64chars[(y << 2) & 63];
	    *p++ = '=';
	    break;
	}
	x = *s++;
	*p++ = b64chars[((y << 2) | ((x >> 6) & 3)) & 63];
	*p++ = b64chars[x & 63];
    }
    *p = '\0';
    
    return t;
}


/*
 * base64dec()
 *
 * Decode the BASE64 encoded string 's'. The returned data should be freed
 * when not used anymore. The length of the decoded data is stored in 'len'.
 * If 's' is not a valid base64 encoded string, this function returns NULL 
 * and sets 'len' to 0.
 */

char *base64dec(char *s, size_t *len)
{
    int x, y;
    char *p;
    char *t;
    
    unsigned char b64_decode[] =
    {
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255, 62,255,255,255, 63,
	 52, 53, 54, 55, 56, 57, 58, 59, 60, 61,255,255,255,255,255,255,
	255,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
	 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25,255,255,255,255,255,
	255, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
	 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51,255,255,255,255,255
    };
    
    t = (char *)xmalloc(3 * (strlen(s) / 4) + 1);
    p = t;
    
    while ((x = (int)((unsigned char)(*s++))) != 0)
    {
	if (x > 127 || (x = b64_decode[x]) == 255)
	{
	    goto error_exit;
	}
	if ((y = (int)((unsigned char)(*s++))) == 0
		|| (y = b64_decode[y]) == 255)
	{
	    goto error_exit;
	}
	*p++ = (char)(x << 2) | (y >> 4);
	if ((x = (int)((unsigned char)(*s++))) == '=') 
	{
	    if (*s++ != '=' || *s != 0)
	    {
		goto error_exit;
	    }
	}
	else 
	{
	    if (x > 127 || (x = b64_decode[x]) == 255)
	    {
		goto error_exit;
	    }
	    *p++ = (char)(y << 4) | (x >> 2);
	    if ((y = (int)((unsigned char)(*s++))) == '=') 
	    {
		if (*s != '\0')
		{
		    goto error_exit;
		}
	    }
	    else 
	    {
		if (y > 127 || (y = b64_decode[y]) == 255)
		{
		    goto error_exit;
		}
		*p++ = (char)(x << 6) | y;
	    }
	}
    }
    *p = '\0';
    *len = p - t;
    return t;

error_exit:
    free(t);
    add_error_line("bad string in base64dec()");
    *len = 0;
    return NULL;
}

#endif /* USE_GSASL */
