/*
 * esmtp.c
 *
 * This file is part of the msmtp package, a very simple programm that 
 * works as an "SMTP plugin" for Mutt and possibly other MUAs.
 *
 * Copyright (C) 2000, 2003	Martin Lambers <marlam@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_VSNPRINTF 
#define _GNU_SOURCE
#include <stdio.h>
#else
#include <stdio.h>
#endif /* HAVE_VSNPRINTF */
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <sysexits.h>

#ifdef USE_GSASL
#include <gsasl.h>
#else
#include "base64.h"
#include "cram_md5.h"
#endif

#include "merror.h"
#include "net.h"
#include "tls.h"
#include "xmalloc.h"
#include "esmtp.h"


/* according to RFC2822, an SMTP line can contain at most 998 
 * characters + CRLF. We add one character for '\0'.
 * If a line is longer, your MUA (in case of writing) or your
 * providers SMTP sever (in case of reading) violates RFC2822.
 */
#define SMTP_BUFSIZE 1001 	

/* SMTP server capabilities, advertised in the EHLO response */
typedef struct
{
    unsigned int flags;
} smtp_cap_t;
#define CAP_STARTTLS		 1
#define CAP_AUTH		 2
#define CAP_AUTH_PLAIN		 4
#define CAP_AUTH_LOGIN		 8
#define CAP_AUTH_CRAM_MD5	16
#define CAP_AUTH_DIGEST_MD5	32
#define CAP_AUTH_NTLM		64


/*
 * smtp_get_full_response()
 * 
 * This function gets one line of a message from an smtp server on 'fd'.
 * If 'tls' is an active TLS connection, TLS will be used.
 * The line will be written to 'buffer' which is expected to be SMTP_BUFSIZE
 * bytes long.
 * Return value is the three digit status number or -1 if an error occured.
 */

int smtp_get_full_response(int fd, tls_t *tls, char *buffer)
{
    int error_code;

    if (tls_is_active(tls))
    {
	error_code = tls_getline(tls, buffer, SMTP_BUFSIZE);
    }
    else
    {
	error_code = net_getline(fd, buffer, SMTP_BUFSIZE);
    }
    if (error_code != 0)
    {
	if (error_code == 1)
	{
	    add_error_line("cannot get a response from SMTP server");
	}
	else /* error_code == 2 */
	{
	    add_error_line("SMTP server sent a line longer than 998 "
		    "characters (this violates RFC 2822)");
	}
	return -1;
    }
    
    /* printf("DEBUG: <-- %s\n", buffer); */
    return atoi(buffer);
}


/*
 * smtp_server_msg_continues()
 *
 * Return 1 if the SMTP server message contains more lines and 0 otherwise.
 */

int smtp_server_msg_continues(char *buffer)
{
    return (buffer[0] && buffer[1] && buffer[2] && buffer[3] == '-');
}


/*
 * smtp_add_error()
 *
 * Adds the SMTP response in 'buffer' to the error messages.
 */

void smtp_add_error(char *buffer)
{
    add_error_line("SMTP error: `%s'", buffer);
}


/*
 * smtp_get_response()
 * 
 * This function gets a message from an smtp server on 'fd'.
 * If 'tls' is an active TLS connection, TLS will be used.
 * Returns the three digit status number. If this number does not equal 
 * 'expected_status', the message from the server is stored via the merror.c
 * functions. Multiline messages are handled. 
 * Return value is the status number or -1 if an error occured.
 */

int smtp_get_response(int fd, tls_t *tls, int expected_status)
{
    int smtp_status;
    char buf[SMTP_BUFSIZE];
    int error = 0;
    

    if ((smtp_status = smtp_get_full_response(fd, tls, buf)) == -1)
    {
	return -1;
    }
    
    if (smtp_status != expected_status)
    {
	smtp_add_error(buf);
	error = 1;
    }
    
    /* read continuation lines */
    while (smtp_server_msg_continues(buf))
    {
	if ((smtp_status = smtp_get_full_response(fd, tls, buf)) == -1)
	{
	    return -1;
	}
	if (error)
	{
	    smtp_add_error(buf);
	}
    }

    return smtp_status;
}
	

/*
 * smtp_put()
 *
 * This function is a printf that writes to an SMTP server on 'fd'.
 * If 'tls' is an active TLS connection, TLS will be used. 
 * The string should not end with '\n', because this function appends 
 * an '\r' and an '\n' to it (TCP CRLF).
 * Return value is 1 on write error or buffer overflow and 0 on success.
 */

int smtp_put(int fd, tls_t *tls, char *format, ...)
{
    va_list args;
    int error_code;
    char line[SMTP_BUFSIZE];

    error_code = 0;
    va_start(args, format);
    if (vsnprintf(line, (size_t)(SMTP_BUFSIZE - 3), format, args) 
	    >= SMTP_BUFSIZE - 3)
    {
	/* should never happen */
	add_error_line("line too long in smtp_put()");
	error_code = 1;
    }
    else
    {
	/* printf("DEBUG: --> %s\n", line); */
	strcat(line, "\r\n");
	if (tls_is_active(tls))
	{
	    error_code = tls_putline(tls, line);
	}
	else
	{
	    error_code = net_putline(fd, line);
	}
	if (error_code != 0)
	{
	    add_error_line("cannot write to SMTP server");
	    error_code = 1;
	}
    }
    va_end(args);
    
    return error_code;
}


/*
 * smtp_ehlo()
 *
 * Sends the EHLO command to the SMTP server and sets 'cap' (based on the
 * response) indicating which features the server supports.
 * Returns a value from sysexits.h
 */

int smtp_ehlo(int fd, tls_t *tls, smtp_cap_t *cap)
{
    char buffer[SMTP_BUFSIZE];

    
    cap->flags = 0;
    if (smtp_put(fd, tls, "EHLO localhost") != 0)
    {
	return EX_IOERR;
    }

    for (;;)
    {
	if (smtp_get_full_response(fd, tls, buffer) == -1)
	{
	    return EX_IOERR;
	}    
	if (strncmp(buffer, "250", 3) != 0 || !buffer[3])
	{
	    smtp_add_error(buffer);
	}
	else
	{
	    if (strncmp(buffer + 4, "STARTTLS", 8) == 0)
	    {
		cap->flags |= CAP_STARTTLS;
	    }
	    else if (strncmp(buffer + 4, "AUTH ", 5) == 0)
	    {
		cap->flags |= CAP_AUTH;
		if (strstr(buffer + 9, "PLAIN"))
		{
		    cap->flags |= CAP_AUTH_PLAIN;
		}
		if (strstr(buffer + 9, "LOGIN"))
		{
		    cap->flags |= CAP_AUTH_LOGIN;
		}
		if (strstr(buffer + 9, "CRAM-MD5"))
		{
		    cap->flags |= CAP_AUTH_CRAM_MD5;
		}
		if (strstr(buffer + 9, "DIGEST-MD5"))
		{
		    cap->flags |= CAP_AUTH_DIGEST_MD5;
		}
		if (strstr(buffer + 9, "NTLM"))
		{
		    cap->flags |= CAP_AUTH_NTLM;
		}
	    }
	}
	if (!smtp_server_msg_continues(buffer))
	{
	    break;
	}
    }

    return EX_OK;
}


/*
 * smtp_starttls()
 *
 * Starts TLS encryption in an SMTP session. 
 * If the 'nostarttls' flag is set, the STARTTLS command will not be sent, TLS
 * will be started immediately.
 * If it is not set, the STARTTLS command will only be sent if 'cap->flags' 
 * contains CAP_STARTTLS. After that, TLS will be started.
 * The 'hostname' and 'tls_trust_file' arguments are needed for certificate
 * verification; see tls_start().
 * The 'tls_key_file' and 'tls_cert_file' arguments are needed if the server 
 * requests a client certificate; see tls_start().
 * All '*_file' may be NULL.
 * Returns a value from sysexits.h
 */

int smtp_starttls(int fd, tls_t *tls, char *hostname, 
	char *tls_key_file, char *tls_cert_file, char *tls_trust_file, 
	smtp_cap_t *cap, int nostarttls)
{
    if (!nostarttls)
    {
	if (!(cap->flags & CAP_STARTTLS))
	{
	    add_error_line("SMTP server does not support "
		    "the STARTTLS command");
	    return EX_UNAVAILABLE;
	}
	if (smtp_put(fd, tls, "STARTTLS") != 0)
	{
	    return EX_IOERR;
	}
	if (smtp_get_response(fd, tls, 220) != 220)
	{
	    add_error_line("line `STARTTLS' not accepted by SMTP server");
	    return EX_UNAVAILABLE;
	}
    }
    if (tls_start(tls, fd, hostname, 
		tls_key_file, tls_cert_file, tls_trust_file) != 0)
    {
	add_error_line("cannot start TLS encryption");
	return EX_UNAVAILABLE;
    }
    
    return 0;
}


/*
 * smtp_auth_plain()
 * 
 * Do SMTP authentication via AUTH PLAIN.
 * Returns a value from sysexits.h
 */

#ifndef USE_GSASL
int smtp_auth_plain(int fd, tls_t *tls, char *user, char *password)
{
    char *s;
    char *b64;
    size_t u_len;
    size_t p_len;
    int error_code;

    u_len = strlen(user);
    p_len = strlen(password);
    s = (char *)xmalloc((u_len + p_len + 3) * sizeof(char));
    s[0] = '\0';
    strcpy(s + 1, user);
    strcpy(s + u_len + 2, password);
    b64 = base64enc(s, 1 + u_len + 1 + p_len);
    free(s);
    
    if (smtp_put(fd, tls, "AUTH PLAIN %s", b64) != 0)
    {
	free(b64);
	return EX_IOERR;
    }
    free(b64);
    if ((error_code = smtp_get_response(fd, tls, 235)) != 235)
    {
	if (error_code == 504)
	{
	    add_error_line("AUTH PLAIN not accepted by SMTP server");
	    return EX_UNAVAILABLE;
	}
	else
	{
	    add_error_line("AUTH PLAIN: SMTP server does not "
		"accept username/password");
	    return EX_DATAERR;
	}
    }

    return 0;
}
#endif


/*
 * smtp_auth_login()
 * 
 * Do SMTP authentication via AUTH LOGIN.
 * Returns a value from sysexits.h
 */

#ifndef USE_GSASL
int smtp_auth_login(int fd, tls_t *tls, char *user, char *password)
{
    char *b64;
    
    if (smtp_put(fd, tls, "AUTH LOGIN") != 0)
    {
	return EX_IOERR;
    }
    if (smtp_get_response(fd, tls, 334) != 334)
    {
	add_error_line("line `AUTH LOGIN' not accepted by SMTP server");
	return EX_UNAVAILABLE;
    }
    b64 = base64enc(user, strlen(user));
    if (smtp_put(fd, tls, "%s", b64) != 0)
    {
	free(b64);
	return EX_IOERR;
    }
    free(b64);
    if (smtp_get_response(fd, tls, 334) != 334)
    {
	add_error_line("AUTH LOGIN: SMTP server does not accept username");
	return EX_DATAERR;
    }
    b64 = base64enc(password, strlen(password));
    if (smtp_put(fd, tls, "%s", b64) != 0)
    {
	free(b64);
	return EX_IOERR;
    }
    free(b64);
    if (smtp_get_response(fd, tls, 235) != 235)
    {
	add_error_line("AUTH LOGIN: SMTP server does not "
		"accept username/password");
	return EX_DATAERR;
    }

    return 0;
}
#endif


/*
 * smtp_auth_cram_md5()
 *
 * Do SMTP authentication via AUTH CRAM-MD5.
 * Returns a value from sysexits.h
 */

#ifndef USE_GSASL
int smtp_auth_cram_md5(int fd, tls_t *tls, char *user, char *password)
{
    char buffer[SMTP_BUFSIZE];
    unsigned char digest[16];
    char hex[16] = 
    	{ '0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f' };
    char *b64buf;
    char *p;
    size_t len;
    int i;
    
    if (smtp_put(fd, tls, "AUTH CRAM-MD5") != 0)
    {
	return EX_IOERR;
    }
    if (smtp_get_full_response(fd, tls, buffer) == -1)
    {
	return EX_IOERR;
    }
    if (strncmp(buffer, "334 ", 4) != 0)
    {
	add_error_line("SMTP error: `%s'", buffer);
	add_error_line("line `AUTH CRAM-MD5' not accepted by SMTP server");
	return EX_UNAVAILABLE;
    }
    b64buf = base64dec(buffer + 4, &len);
    if (!b64buf)
    {
	/* should never happen */
	add_error_line("AUTH CRAM-MD5: server sent invalid challenge");
	return EX_PROTOCOL;
    }
    cram_md5(password, strlen(password), b64buf, len, digest);
    free(b64buf);
    /* need enough space for base64(username + ' ' + digest_in_hex + '\0') */
    if (4 * ((strlen(user) + 1 + 32 + 1 + 2) / 3) + 1 > SMTP_BUFSIZE)
    {
	/* should never happen */
	add_error_line("username too long for AUTH CRAM-MD5 authentication");
	return EX_DATAERR;
    }
    /* buffer is long enough */
    strcpy(buffer, user);
    p = buffer + strlen(user);
    *p++ = ' ';
    for (i = 0; i < 16; i++)
    {
	p[2 * i] = hex[digest[i] >> 4];
	p[2 * i + 1] = hex[digest[i] & 0x0f];
    }
    p[32] = '\0';
    p = base64enc(buffer, strlen(buffer));
    /* buffer is long enough */
    strcpy(buffer, p);
    free(p);
    if (smtp_put(fd, tls, "%s", buffer) != 0)
    {
	return EX_IOERR;
    }
    if (smtp_get_response(fd, tls, 235) != 235)
    {
	add_error_line("AUTH CRAM-MD5: SMTP server does not "
		"accept username/password");
	return EX_DATAERR;
    }

    return 0;
}
#endif


/*
 * GSASL callbacks for authentication
 */

#ifdef USE_GSASL

char *auth_username;
char *auth_password;
char *auth_hostname;
char *auth_servicename;

int callback_username(Gsasl_session_ctx *ctx, char *out, size_t *outlen)
{
    *outlen = strlen(auth_username);
    if (out)
    {
	strcpy(out, auth_username);
    }
    return GSASL_OK;
}

int callback_password(Gsasl_session_ctx *ctx, char *out, size_t *outlen)
{
    *outlen = strlen(auth_password);
    if (out)
    {
	strcpy(out, auth_password);
    }
    return GSASL_OK;
}

int callback_service(Gsasl_session_ctx *ctx, 
	char *service, size_t *servicelen, 
	char *hostname, size_t *hostnamelen, 
	char *servicename, size_t *servicenamelen)
{
    *servicelen = 4;
    *hostnamelen = strlen(auth_hostname);
    *servicenamelen = 0;
    if (service)
    {
	strcpy(service, "smtp");
    }
    if (hostname)
    {
	strcpy(hostname, auth_hostname);
    }
    if (servicename)
    {
	strcpy(servicename, "");
    }
    return GSASL_OK;
}
#endif


/*
 * smtp_auth()
 *
 * Do SMTP authentication with the specified method 'auth_mech'.
 * If 'auth_mech' is "", choose the best mechanism available that does not
 * send cleartext passwords.
 * Returns a value from sysexits.h
 */

int smtp_auth(int fd, tls_t *tls, char *auth_mech, char *user, char *password,
	char *hostname, smtp_cap_t *cap, int use_tls)
{
    int error_code;
#ifdef USE_GSASL
    Gsasl_ctx *ctx;
    Gsasl_session_ctx *cctx;
    char *input;
    char inbuf[SMTP_BUFSIZE];
    char outbuf[SMTP_BUFSIZE];
#endif

    if (!(cap->flags & CAP_AUTH))
    {
	add_error_line("SMTP server does not support authentication");
	return EX_UNAVAILABLE;
    }
    /* Choose "best" authentication mechanism. Do not use
     * gsasl_client_suggest_mechanism() because it seems to always return
     * "ANONYMOUS" (libgsasl 0.0.7) */
    if (strcmp(auth_mech, "") == 0)
    {
#ifdef USE_GSASL
	if (cap->flags & CAP_AUTH_DIGEST_MD5)
	{
	    auth_mech = "DIGEST-MD5";
	}
	else
#endif
	if (cap->flags & CAP_AUTH_CRAM_MD5)
	{
	    auth_mech = "CRAM-MD5";
	}
#ifdef USE_GSASL
	else if (cap->flags & CAP_AUTH_NTLM)
	{
	    auth_mech = "NTLM";
	}
#endif
	else if (use_tls)
	{
	    if (cap->flags & CAP_AUTH_PLAIN)
	    {
		auth_mech = "PLAIN";
	    }
	    else if (cap->flags & CAP_AUTH_LOGIN)
	    {
		auth_mech = "LOGIN";
	    }
	    else
	    {
		add_error_line("SMTP server does not support an "
			"authentication method that I know of");
		return EX_UNAVAILABLE;
	    }
	}
	else
	{
	    add_error_line("SMTP server does not support a secure "
		    "authentication method that I know of");
	    add_error_line("Please enable TLS encryption (recommended) "
		    "or choose an insecure method yourself");
	    return EX_UNAVAILABLE;
	}
    }
    
#ifdef USE_GSASL
    
    if ((error_code = gsasl_init(&ctx)) != GSASL_OK)
    {
	add_error_line("Cannot initialize GSASL authentication library "
		"[%s]", gsasl_strerror(error_code));
	return EX_SOFTWARE;
    }
    auth_username = user;
    auth_password = password;
    auth_hostname = hostname;
    gsasl_client_callback_authentication_id_set(ctx, callback_username);
    gsasl_client_callback_authorization_id_set(ctx, callback_username);
    gsasl_client_callback_password_set(ctx, callback_password);
    gsasl_client_callback_service_set(ctx, callback_service);
    if ((error_code = gsasl_client_start(ctx, auth_mech, &cctx)) != GSASL_OK)
    {
	add_error_line("Cannot start authentication mechanism `%s' [%s]", 
		auth_mech, gsasl_strerror(error_code));
	gsasl_done(ctx);
	return EX_SOFTWARE;
    }    
    input = NULL;
    do
    {
	error_code = gsasl_client_step_base64(
     		cctx, input, outbuf, SMTP_BUFSIZE);
	/* AUTH PLAIN needs special treatment because it needs to send the
	 * authentication data together with the AUTH PLAIN command */
	if (!input && strcmp(auth_mech, "PLAIN") == 0 && outbuf[0])
	{
	    if (smtp_put(fd, tls, "AUTH PLAIN %s", outbuf) != 0)
	    {
		gsasl_done(ctx);
		return EX_IOERR;
	    }
	    if (smtp_get_full_response(fd, tls, inbuf) == -1)
	    {
		gsasl_done(ctx);
		return EX_IOERR;
	    }
	    if (strncmp(inbuf, "334 ", 4) != 0 && strncmp(inbuf, "235 ", 4) != 0)
	    {
		add_error_line("Authentication error");
		gsasl_done(ctx);
		return EX_PROTOCOL;
	    }	    
	    input = inbuf + 4;
	    continue;
	}
	if (!input)
	{
	    if (smtp_put(fd, tls, "AUTH %s", auth_mech) != 0)
	    {
		gsasl_done(ctx);
		return EX_IOERR;
	    }
	    if (smtp_get_full_response(fd, tls, inbuf) == -1)
	    {
		gsasl_done(ctx);
		return EX_IOERR;
	    }
	    if (strncmp(inbuf, "334 ", 4) != 0 && strncmp(inbuf, "235 ", 4) != 0)
	    {
		add_error_line("Authentication error");
		gsasl_done(ctx);
		return EX_PROTOCOL;
	    }	    
	    input = inbuf + 4;
	}
	if (outbuf[0])
	{
	    if (smtp_put(fd, tls, "%s", outbuf) != 0)
	    {
		gsasl_done(ctx);
		return EX_IOERR;
	    }
	    if (smtp_get_full_response(fd, tls, inbuf) == -1)
	    {
		gsasl_done(ctx);
		return EX_IOERR;
	    }
	    if (strncmp(inbuf, "334 ", 4) != 0 && strncmp(inbuf, "235 ", 4) != 0)
	    {
		add_error_line("Authentication error");
		gsasl_done(ctx);
		return EX_PROTOCOL;
	    }	    
	    input = inbuf + 4;
	}
    }
    while (error_code == GSASL_NEEDS_MORE);
    if (error_code != GSASL_OK)
    {
	add_error_line("Authentication failed [%s]", 
		gsasl_strerror(error_code));
	gsasl_done(ctx);
	return EX_DATAERR;
    }
    /* In some cases (ok, only DIGEST-MD5), we need to send an empty
     * answer to the last 334 response before we get 235. */
    if (strncmp(inbuf, "235 ", 4) != 0)
    {
	if (smtp_put(fd, tls, "") != 0)
	{
    	    gsasl_done(ctx);
	    return EX_IOERR;
	}
	if (smtp_get_response(fd, tls, 235) != 235)
	{
	    add_error_line("Authentication failed [%s]", 
	    	    gsasl_strerror(error_code));
	    gsasl_done(ctx);
	    return EX_DATAERR;
	}
    }
    gsasl_done(ctx);
    return 0;

#else
    
    if (strcmp(auth_mech, "CRAM-MD5") == 0)
    {
	if ((error_code = smtp_auth_cram_md5(fd, tls, user, password)) != 0)
	{
	    add_error_line("AUTH CRAM-MD5 authentication failed");
	    return error_code;
	}
    }
    else if (strcmp(auth_mech, "PLAIN") == 0)
    {
	if ((error_code = smtp_auth_plain(fd, tls, user, password)) != 0)
	{
	    add_error_line("AUTH PLAIN authentication failed");
	    return error_code;
	}
    }
    else if (strcmp(auth_mech, "LOGIN") == 0)
    {
	if ((error_code = smtp_auth_login(fd, tls, user, password)) != 0)
	{
	    add_error_line("AUTH LOGIN authentication failed");
	    return error_code;
	}
    }
    else
    {
	add_error_line("BUG: unknown SMTP AUTH method in smtp_auth()");
	return EX_SOFTWARE;
    }

    return 0;
#endif
}


/*
 * smtp_send()
 *
 * Send the mail (read from 'mailf') to the SMTP server on 'server:port'.
 * Authenticate with 'auth_mech' (NULL means no authentication, "" means choose
 * mechanism). 'user' and 'password' must not be NULL if 'auth_mech' != NULL).
 * Use TLS encryption if 'use_tls' is set.
 * The number of recipients is stored in 'rcptc', the recipient mail addresses 
 * in 'rcptv'. The 'from' string should contain the mail address of the user to
 * be accepted by the server. 
 * The Bcc header will be removed from the mail and not sent to the SMTP
 * server unless the 'keep_bcc' flag is set.
 * If 'use_tls' is set, TLS will be activated using 'tls_key_file', 'tls_cert_file',
 * 'tls_trust_file' and 'nostarttls' (see smtp_starttls()).
 * The function returns 0 if the message was successfully send, and a value
 * from sysexits.h indicating the error on failure.
 */

int smtp_send(char *server, int port, char *from, 
	char *auth_mech, char *user, char *password,
	int use_tls, char *tls_key_file, char *tls_cert_file, 
	char *tls_trust_file, int nostarttls, 
	int keep_bcc,
	int rcptc, char *rcptv[], 
	FILE *mailf)
{
    smtp_cap_t cap;
    char buffer[SMTP_BUFSIZE];
    char *p;
    int error_code;
    int i;
    int fd;
    tls_t tls;
    int in_header;
    int in_bcc;
    

    tls_clear(&tls);
    /* connect to smtp server */
    if ((fd = open_socket(server, port)) < 0)
    {
	add_error_line("cannot connect to SMTP server");
	error_code = EX_UNAVAILABLE;
	goto exit;
    }
    /* start TLS encryption for TLS-only servers (no STARTTLS) if wanted */
    if (use_tls && nostarttls)
    {
	if ((error_code = smtp_starttls(fd, &tls, server, tls_key_file, 
			tls_cert_file, tls_trust_file, &cap, nostarttls)) != 0)
	{
	    add_error_line("TLS encryption failed");
	    goto close_exit;
	}
    }
    /* get initial ok message */
    if (smtp_get_response(fd, &tls, 220) != 220)
    {
	add_error_line("cannot get initial ok message from SMTP server");
	error_code = EX_UNAVAILABLE;
	goto close_exit;
    }
    /* start smtp session */
    if ((error_code = smtp_ehlo(fd, &tls, &cap)) != EX_OK)
    {
	goto close_exit;
    }
    /* start TLS encryption with STARTTLS if wanted */
    if (use_tls && !nostarttls)
    {
	if ((error_code = smtp_starttls(fd, &tls, server, tls_key_file,
			tls_cert_file, tls_trust_file, &cap, nostarttls)) != 0)
	{
	    add_error_line("TLS encryption failed");
	    goto close_exit;
	}
	/* resend EHLO command and parse the response again after successful
	 * TLS handshake (this is a SHOULD clause in RFC 2487 and required 
	 * by some MTAs) */
	if ((error_code = smtp_ehlo(fd, &tls, &cap)) != EX_OK)
	{
	    goto close_exit;
	}
    }
    /* do SMTP AUTH authentication if wanted */
    if (auth_mech)
    {
	if ((error_code = smtp_auth(fd, &tls, auth_mech, 
			user, password, server, &cap, use_tls)) != EX_OK)
	{
	    goto close_exit;
	}
    }
    /* tell the server who wants to send mail and to whom */
    if (smtp_put(fd, &tls, "MAIL FROM:<%s>", from) != 0)
    {
	error_code = EX_IOERR;
	goto close_exit;
    }
    if (smtp_get_response(fd, &tls, 250) != 250)
    {
	add_error_line("line `MAIL FROM:<%s>' not accepted by SMTP server", 
		from);
	error_code = EX_DATAERR;
	goto close_exit;
    }
    for (i = 0; i < rcptc; i++)
    {
	if (smtp_put(fd, &tls, "RCPT TO:<%s>", rcptv[i]) != 0)
	{
	    error_code = EX_IOERR;
	    goto close_exit;
	}
	if (smtp_get_response(fd, &tls, 250) != 250)
	{
	    add_error_line("line `RCPT TO:<%s>' not accepted by SMTP server", 
		    rcptv[i]);
	    error_code = EX_DATAERR;
	    goto close_exit;
	}
    }
    /* send the mail */
    if (smtp_put(fd, &tls, "DATA") != 0)
    {
	error_code = EX_IOERR;
	goto close_exit;
    }
    if (smtp_get_response(fd, &tls, 354) != 354)
    {
	add_error_line("SMTP server does not accept the mail");
	error_code = EX_DATAERR;
	goto close_exit;
    }
    /* Copy the mail from mailf to fd, converting '\n' to '\r\n' and a leading
     * dot on a line to two leading dots; the smtp server deletes a leading
     * dot (RFC 2821, section 4.5.2).
     * Strip the Bcc header unless the keep_bcc flag is set. */
    in_header = 1;
    in_bcc = 0;
    while (fgets(buffer, (int)sizeof(buffer), mailf) != NULL)
    {
	p = strchr(buffer, '\n');
	if (!p)
	{
	    add_error_line("line longer than 998 characters; "
		    "your MUA violates RFC 2822!");
	    error_code = EX_DATAERR;
	    goto close_exit;
	}
	*p = '\0';
	if (!keep_bcc)
	{
	    if (in_header && buffer[0] == '\0')
	    {
		in_header = 0;
	    }
	    if (in_header)
	    {
		if (!in_bcc)
		{
		    if (strlen(buffer) >= 5
			    && strncmp(buffer, "Bcc: ", (size_t)5) == 0)
		    {
			in_bcc = 1;
			/* remove Bcc header by ignoring this line */
			continue;
		    }
		}
		else
		{
		    /* continued header lines begin with "horizontal whitespace"
		     * (RFC 2822, section 2.2.3) */
		    if (buffer[0] == '\t' || buffer[0] == ' ')
		    {
			/* remove continued Bcc header by ignoring this line */
			continue;
		    }
		    else
		    {
			in_bcc = 0;
		    }
		}
	    }
	}
	if (buffer[0] == '.')
	{
	    /* This works because we turned '\n\0' to '\0\0'. */
	    memmove(buffer + 1, buffer, strlen(buffer) * sizeof(char));
	    buffer[0] = '.';
	}
	/* smtp_put will append "\r\n" */
	if (smtp_put(fd, &tls, "%s", buffer) != 0)
	{
	    error_code = EX_IOERR;
	    goto close_exit;
	}
    }
    if (ferror(mailf))
    {
	add_error_line("input error while reading the mail");
	error_code = EX_IOERR;
	goto close_exit;
    }
    /* end the mail */
    if (smtp_put(fd, &tls, ".") != 0)
    {
	error_code = EX_IOERR;
	goto close_exit;
    }
    if (smtp_get_response(fd, &tls, 250) != 250)
    {
	add_error_line("SMTP server did not accept the mail");
	error_code = EX_DATAERR;
	goto close_exit;
    }
    
    /* Since the smtp server responded with 250, we can be sure that the mail
     * will be delivered. Our job is done. We can ignore following i/o and
     * smtp errors. */
    error_code = 0;
    smtp_put(fd, &tls, "QUIT");
    smtp_get_response(fd, &tls, 221);

close_exit:
    /* close connection and clean up */
    tls_close(&tls);
    close(fd);

exit:
    return error_code;
}
